@@ -4,9 +4,14 @@ import numpy as np
 import pandas as pd
 import pytest
 
-from pandas.testing import assert_index_equal, assert_series_equal
+from pandas.testing import (
+    assert_index_equal,
+    assert_series_equal,
+    assert_frame_equal
+)
 
 from solarforecastarbiter import datamodel
+from solarforecastarbiter.conftest import _site_metadata
 from solarforecastarbiter.metrics import preprocessing
 
 
@@ -421,6 +426,7 @@ def test_align_prob_constant_value(
                 [1.], index=pd.DatetimeIndex(["20200301T03Z"], freq='1h')),
             (('ISNAN', 1), ('NIGHTTIME', 2), ('USER FLAGGED', 2),
              ('TOTAL DISCARD BEFORE RESAMPLE', 3),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
         ),
@@ -432,6 +438,7 @@ def test_align_prob_constant_value(
                 [1.], index=pd.DatetimeIndex(["20200301T03Z"], freq='1h')),
             (('ISNAN', 1), ('NIGHTTIME', 2), ('USER FLAGGED', 2),
              ('TOTAL DISCARD BEFORE RESAMPLE', 3),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR ISNAN', 2, False),
              ('USER FLAGGED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
@@ -445,6 +452,7 @@ def test_align_prob_constant_value(
                 [14.5], index=pd.DatetimeIndex(["20200301T03Z"], freq='1h')),
             (('ISNAN', 5), ('NIGHTTIME', 3), ('USER FLAGGED', 3),
              ('TOTAL DISCARD BEFORE RESAMPLE', 8),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
         ),
@@ -457,6 +465,7 @@ def test_align_prob_constant_value(
             pd.Series([7.5, 11.5, 14.5], index=FOUR_HOUR_SERIES.index[1:]),
             (('ISNAN', 5), ('NIGHTTIME', 3), ('USER FLAGGED', 3),
              ('TOTAL DISCARD BEFORE RESAMPLE', 8),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 1, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 1, False))
         ),
@@ -468,6 +477,7 @@ def test_align_prob_constant_value(
                 [14.5], index=pd.DatetimeIndex(["20200301T03Z"], freq='1h')),
             (('ISNAN', 5), ('NIGHTTIME', 3), ('USER FLAGGED', 3),
              ('TOTAL DISCARD BEFORE RESAMPLE', 8),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR ISNAN', 3, False),
              ('USER FLAGGED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
@@ -482,6 +492,7 @@ def test_align_prob_constant_value(
                 [14.5], index=pd.DatetimeIndex(["20200301T03Z"], freq='1h')),
             (('ISNAN', 5),
              ('TOTAL DISCARD BEFORE RESAMPLE', 5),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
         ),
@@ -498,6 +509,7 @@ def test_align_prob_constant_value(
             pd.Series([10.5, 14.5], index=FOUR_HOUR_SERIES.index[2:]),
             (('ISNAN', 5),
              ('TOTAL DISCARD BEFORE RESAMPLE', 5),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR ISNAN', 2, False),
              ('USER FLAGGED OR ISNAN', 2, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 2, False))
@@ -511,6 +523,7 @@ def test_align_prob_constant_value(
             # only first all-nan interval is discarded
             pd.Series([7.0, 10.5, 14.5], index=FOUR_HOUR_SERIES.index[1:]),
             (('ISNAN', 5), ('TOTAL DISCARD BEFORE RESAMPLE', 5),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 1, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 1, False))
         ),
@@ -532,6 +545,7 @@ def test_align_prob_constant_value(
                 ),
             (('ISNAN', 5), ('CLEARSKY EXCEEDED', 1),
              ('TOTAL DISCARD BEFORE RESAMPLE', 6),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR USER FLAGGED OR ISNAN', 1, False),
              ('CLEARSKY EXCEEDED OR ISNAN', 3, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 3, False))
@@ -553,6 +567,7 @@ def test_align_prob_constant_value(
                 )
             ),
             (('ISNAN', 0), ('TOTAL DISCARD BEFORE RESAMPLE', 0),
+             ('OUTAGE', 0),
              ('NIGHTTIME OR ISNAN', 1, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 1, False))
         ),
@@ -571,6 +586,7 @@ def test_align_prob_constant_value(
                 )
             ),
             (('ISNAN', 0), ('TOTAL DISCARD BEFORE RESAMPLE', 0),
+             ('OUTAGE', 0),
              ('USER FLAGGED OR ISNAN', 1, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 1, False))
         ),
@@ -591,6 +607,7 @@ def test_align_prob_constant_value(
             (('ISNAN', 0),
              ('USER FLAGGED', 5, True),
              ('TOTAL DISCARD BEFORE RESAMPLE', 5, True),
+             ('OUTAGE', 0),
              ('USER FLAGGED OR ISNAN', 1, False),
              ('TOTAL DISCARD AFTER RESAMPLE', 1, False))
         )
@@ -681,6 +698,62 @@ def test_process_forecast_observations(report_objects, quality_filter,
                                       proc_fxobs.observation_values.index)
 
 
+def test_process_forecast_observations_with_outages(
+    report_objects_with_outages, quality_filter,
+    timeofdayfilter, mocker
+):
+    report, observation, forecast_0, forecast_1, aggregate, forecast_agg = report_objects_with_outages  # NOQA: E501
+    forecast_ref = report.report_parameters.object_pairs[1].reference_forecast
+    obs_ser = pd.Series(np.arange(8),
+                        index=pd.date_range(start='2019-04-01T00:00:00',
+                                            periods=8,
+                                            freq='15min',
+                                            tz='MST',
+                                            name='timestamp'))
+    obs_df = obs_ser.to_frame('value')
+    obs_df['quality_flag'] = OK
+    agg_df = THREE_HOUR_SERIES.to_frame('value')
+    agg_df['quality_flag'] = OK
+    data = {
+        observation: obs_df,
+        forecast_0: THREE_HOUR_SERIES,
+        forecast_1: THREE_HOUR_SERIES,
+        forecast_ref: THREE_HOUR_SERIES,
+        forecast_agg: THREE_HOUR_SERIES,
+        aggregate: agg_df
+    }
+    filters = [quality_filter, timeofdayfilter]
+    logger = mocker.patch('solarforecastarbiter.metrics.preprocessing.logger')
+    processed_fxobs_list = preprocessing.process_forecast_observations(
+        report.report_parameters.object_pairs,
+        filters,
+        report.report_parameters.forecast_fill_method,
+        report.report_parameters.start,
+        report.report_parameters.end,
+        data, 'MST',
+        costs=report.report_parameters.costs,
+        outages=report.outages)
+    assert len(processed_fxobs_list) == len(
+        report.report_parameters.object_pairs)
+    assert logger.warning.called
+    assert not logger.error.called
+
+    for proc_fxobs in processed_fxobs_list:
+        assert isinstance(proc_fxobs, datamodel.ProcessedForecastObservation)
+        assert all(isinstance(vr, datamodel.ValidationResult)
+                   for vr in proc_fxobs.validation_results)
+        assert all(isinstance(pr, datamodel.PreprocessingResult)
+                   for pr in proc_fxobs.preprocessing_results)
+        assert isinstance(proc_fxobs.forecast_values, pd.Series)
+        assert isinstance(proc_fxobs.observation_values, pd.Series)
+        pd.testing.assert_index_equal(proc_fxobs.forecast_values.index,
+                                      proc_fxobs.observation_values.index)
+        if proc_fxobs.original.forecast.name == "0 Day GFS GHI":
+            assert proc_fxobs.forecast_values.empty
+        else:
+            assert not proc_fxobs.forecast_values.empty
+
+
 def test_process_probabilistic_forecast_observations(
         cdf_and_cv_report_objects, cdf_and_cv_report_data, quality_filter,
         timeofdayfilter, mocker):
@@ -1653,3 +1726,268 @@ def test_apply_fill_unstratified_dataframe(data, method, exp, exp_count,
                                              data.index[0], data.index[-1])
     pd.testing.assert_frame_equal(result, exp)
     assert count == exp_count
+
+
+OUTAGE_FORECAST = datamodel.Forecast(
+   site=_site_metadata(), name='dummy fx', variable='ghi',
+   interval_value_type='instantaneous',
+   interval_length=pd.Timedelta('1h'),
+   interval_label='beginning',
+   issue_time_of_day=dt.time(hour=5),
+   lead_time_to_start=pd.Timedelta('1h'),
+   run_length=pd.Timedelta('12h')
+)
+
+
+@pytest.mark.parametrize("forecast,start,end,expected", [
+    (OUTAGE_FORECAST,
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     pd.date_range(
+        '2020-12-31T17:00Z',
+        '2021-01-01T17:00Z',
+        freq='12h'
+     )
+     ),
+    (OUTAGE_FORECAST.replace(run_length=pd.Timedelta('1H')),
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     pd.date_range(
+        '2020-12-31T23:00Z',
+        '2021-01-01T22:00Z',
+        freq='1h'
+     )
+     ),
+    (OUTAGE_FORECAST.replace(run_length=pd.Timedelta('24H')),
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     pd.date_range(
+        '2020-12-31T05:00Z',
+        '2021-01-01T05:00Z',
+        freq='24h'
+     )
+     ),
+    (OUTAGE_FORECAST,
+     pd.Timestamp('2021-01-01T04:37Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     pd.date_range(
+        '2020-12-31T17:00Z',
+        '2021-01-01T17:00Z',
+        freq='12h'
+     )
+     ),
+    (OUTAGE_FORECAST.replace(
+        lead_time_to_start=pd.Timedelta('26H'),
+        run_length=pd.Timedelta('24H'),
+        issue_time_of_day=dt.time(hour=23),
+     ),
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     pd.date_range(
+        '2020-12-29T23:00Z',
+        '2020-12-30T23:00Z',
+        freq='24h'
+     )
+     ),
+    (OUTAGE_FORECAST.replace(
+        lead_time_to_start=pd.Timedelta('26H'),
+        run_length=pd.Timedelta('24H'),
+        issue_time_of_day=dt.time(hour=23),
+     ),
+     pd.Timestamp('2021-01-01T00:00', tz='America/Phoenix'),
+     pd.Timestamp('2021-01-02T00:00', tz='America/Phoenix'),
+     pd.date_range(
+        '2020-12-30T23:00Z',
+        '2020-12-31T23:00Z',
+        freq='24h'
+     )
+     ),
+])
+def test_forecast_report_issue_times(forecast, start, end, expected):
+    issue_times = preprocessing.forecast_report_issue_times(
+        forecast, start, end
+    )
+    assert_index_equal(issue_times, expected)
+
+
+@pytest.mark.parametrize("forecast,start,end,outages,expected", [
+    (OUTAGE_FORECAST,
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     (
+        datamodel.TimePeriod(
+            start=pd.Timestamp('2021-01-01T04:00Z'),
+            end=pd.Timestamp('2021-01-01T05:00Z')
+        ),
+      ),
+     (
+        datamodel.TimePeriod(
+            start=pd.Timestamp('2021-01-01T06:00Z'),
+            end=pd.Timestamp('2021-01-01T18:00Z')
+        ),
+      ),
+     ),
+    (OUTAGE_FORECAST,
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     (
+        datamodel.TimePeriod(
+           start=pd.Timestamp('2020-12-31T17:00Z'),
+           end=pd.Timestamp('2020-12-31T19:00Z')
+         ),
+     ),
+     (
+        datamodel.TimePeriod(
+           start=pd.Timestamp('2020-12-31T18:00Z'),
+           end=pd.Timestamp('2021-01-01T06:00Z')
+        ),
+      )
+     ),
+    (OUTAGE_FORECAST.replace(run_length=pd.Timedelta('1H')),
+     pd.Timestamp('2021-01-01T00:00Z'),
+     pd.Timestamp('2021-01-02T00:00Z'),
+     (
+        datamodel.TimePeriod(
+          start=pd.Timestamp('2021-01-01T04:00Z'),
+          end=pd.Timestamp('2021-01-01T05:00Z')
+        ),
+     ),
+     (
+        datamodel.TimePeriod(
+          start=pd.Timestamp('2021-01-01T05:00Z'),
+          end=pd.Timestamp('2021-01-01T06:00Z')
+        ),
+        datamodel.TimePeriod(
+          start=pd.Timestamp('2021-01-01T06:00Z'),
+          end=pd.Timestamp('2021-01-01T07:00Z')
+        ),
+      )
+     )
+])
+def test_outage_periods(forecast, start, end, outages, expected):
+    outage_periods = preprocessing.outage_periods(
+        forecast, start, end, outages
+    )
+    assert outage_periods == expected
+
+
+OUTAGE_DATA = pd.DataFrame(
+    index=pd.date_range(
+        '2021-01-01T05:00Z',
+        '2021-01-02T05:00Z',
+        freq='1H'
+    ),
+    columns=['value']
+)
+
+
+@pytest.mark.parametrize(
+    'outage_periods,interval_label,data,expected_data,dropped',
+    [
+     (
+      [
+         datamodel.TimePeriod(
+           start=pd.Timestamp('2021-01-01T12:00Z'),
+           end=pd.Timestamp('2021-01-01T14:00Z')
+         )
+       ],
+      "beginning",
+      OUTAGE_DATA,
+      pd.DataFrame(
+         index=pd.date_range(
+             '2021-01-01T05:00Z',
+             '2021-01-01T11:00Z',
+             freq='1H'
+         ).append(pd.date_range(
+             '2021-01-01T14:00Z',
+             '2021-01-02T05:00Z',
+             freq='1H'
+         )),
+         columns=['value']
+      ),
+      3
+     ),
+     (
+      [
+         datamodel.TimePeriod(
+           start=pd.Timestamp('2021-01-01T07:00Z'),
+           end=pd.Timestamp('2021-01-01T09:00Z')
+         ),
+         datamodel.TimePeriod(
+           start=pd.Timestamp('2021-01-01T13:00Z'),
+           end=pd.Timestamp('2021-01-01T17:00Z')
+         )
+       ],
+      "beginning",
+      OUTAGE_DATA,
+      pd.DataFrame(
+         index=pd.date_range(
+             '2021-01-01T05:00Z',
+             '2021-01-01T06:00Z',
+             freq='1H'
+         ).append(pd.date_range(
+             '2021-01-01T09:00Z',
+             '2021-01-01T12:00Z',
+             freq='1H'
+         )).append(pd.date_range(
+             '2021-01-01T17:00Z',
+             '2021-01-02T05:00Z',
+             freq='1H'
+         )),
+         columns=['value']
+      ),
+      6
+     ),
+     (
+      [],
+      "beginning",
+      OUTAGE_DATA,
+      OUTAGE_DATA,
+      0
+     ),
+     (
+      [],
+      "ending",
+      OUTAGE_DATA,
+      OUTAGE_DATA,
+      0
+     ),
+     (
+      [
+         datamodel.TimePeriod(
+           start=pd.Timestamp('2021-01-01T07:00Z'),
+           end=pd.Timestamp('2021-01-01T09:00Z')
+         ),
+         datamodel.TimePeriod(
+           start=pd.Timestamp('2021-01-01T13:00Z'),
+           end=pd.Timestamp('2021-01-01T17:00Z')
+         )
+       ],
+      "ending",
+      OUTAGE_DATA,
+      pd.DataFrame(
+         index=pd.date_range(
+             '2021-01-01T05:00Z',
+             '2021-01-01T07:00Z',
+             freq='1H'
+         ).append(pd.date_range(
+             '2021-01-01T10:00Z',
+             '2021-01-01T13:00Z',
+             freq='1H'
+         )).append(pd.date_range(
+             '2021-01-01T18:00Z',
+             '2021-01-02T05:00Z',
+             freq='1H'
+         )),
+         columns=['value']
+      ),
+      6
+     ),
+    ]
+)
+def test_remove_outage_period(
+        outage_periods, interval_label, data, expected_data, dropped):
+    data_outage_removed, actual_dropped = preprocessing.remove_outage_periods(
+        outage_periods, data, interval_label
+    )
+    assert_frame_equal(data_outage_removed, expected_data)
